/////////////////////////////////////////////////////////////////////////////
///
/// @file mwattr.cpp
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2009 by Pawel Pilarczyk.
//
// This file is part of my research program package.  This is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// Started on December 12, 2007. Last revision: December 13, 2007.


// the interval arithmetic package from the CAPD library
#include "capd/interval/DoubleInterval.h"

// the multi-work subdivision scheme from the CHomP library
#include "chomp/multiwork/mwsubdiv.h"

// other files from the CHomP library
#include "chomp/cubes/pointset.h"
#include "chomp/cubes/cube.h"
#include "chomp/struct/digraph.h"


// --------------------------------------------------
// -------------- title and help info ---------------
// --------------------------------------------------

const char *title = "\
MW-Attractors, ver. 0.01. Copyright (C) 1997-2009 by Pawel Pilarczyk.\n\
This is free software. No warranty. Consult 'license.txt' for details.";

const char *helpinfo = "\
This is a program for computing a rigorous inner bound for the set of\n\
parameters of a dynamical system for which there exist at least two\n\
attractors. By default, the program runs for the Leslie population model,\n\
as described in the paper \"Parallelization method for a continuous\n\
property\" by Pawel Pilarczyk. Please, see the paper for more details.";


// --------------------------------------------------
// --------------- Multi-Work options ---------------
// --------------------------------------------------

/// The default port number for the multiwork communication.
/// Use a TCP port available on your computer. Note that
/// numbers below 1024 are reserved for the system. Suggested
/// value is between 4000 and 5000, but anything above 1024
/// and below some 50000 should be fine.
const int defaultPortNumber = 4923;

/// The control number used to confirm that the worker and the
/// coordinator are doing the same computations.
const int controlNumber = 29188;


// --------------------------------------------------
// ------------ the range of parameters -------------
// --------------------------------------------------

/// The dimension of the parameter space. Must not exceed 30.
const int paramDim = 2;

/// The leftmost corner of the entire range of parameters to scan.
const double leftBounds [paramDim] = {10, 10};

/// The rightmost corner of the entire range of parameters to scan.
const double rightBounds [paramDim] = {35, 35};

/// The minimum subdivision level.
const int minSubdivLevel = 2;


// --------------------------------------------------
// ---------------- the phase space -----------------
// --------------------------------------------------

/// The dimension of the phase space.
const int spaceDim = 2;

/// The offset of the lower left corner of the rectangular box R.
double offset [spaceDim] = {-1e-006, -1e-006};

/// The width of the rectangular box R.
double width [spaceDim] = {258, 181};

/// The initial subdivision level to be used in the phase space.
int iniLevel = 5;

/// The final subdivision level to be used in the phase space.
int subLevel = 12;


// --------------------------------------------------
// ---------------- one computation -----------------
// --------------------------------------------------

/// Computes the absolute value of an interval.
template <class IntvT>
inline IntvT abs_value (const IntvT &x)
{
	typename IntvT::BoundType left = x. leftBound ();
	typename IntvT::BoundType right = x. rightBound ();
	if (left >= 0)
		return x;
	else if (right <= 0)
		return IntvT (-right, -left);
	else if ((left <= 0) && (right >= 0))
		return (-left > right) ? IntvT (0.0, -left) :
			IntvT (0.0, right);
	else
		throw "Wrong interval detected in the function 'abs_value'.";
} /* abs_value */

/// Computes the overcompensatory Leslie model function in R^n.
/// Note: In order to make sure that all the input variables are non-negative
/// computes their absolute values before plugging them into the formulas.
template <class Vect1, class Vect2, class Vect3, class Vect4, class CoordT>
inline void Leslie (const Vect1 &x, const Vect2 &theta, const Vect3 &p,
	const CoordT &lambda, Vect4 &result, int dim)
{
	CoordT first = theta [0] * abs_value (x [0]);
	for (int i = 1; i < dim; ++ i)
		first += theta [i] * abs_value (x [i]);
	CoordT sum = abs_value (x [0]);
	for (int i = 1; i < dim; ++ i)
		sum += abs_value (x [i]);
	first *= exp (-lambda * sum);

	for (int i = dim - 1; i > 0; -- i)
		result [i] = p [i - 1] * abs_value (x [i - 1]);
	result [0] = first;

	return;
} /* Leslie */

/// Computes an outer cover of the image of a box in the phase space
/// by the Leslie model map.
/// Input arguments: 'coord' - the integer coordinates of the box,
/// 'dim' - the dimension of the box, 'level' - the subdivision level,
/// 'theta', 'lambda', 'p' - parameters of the Leslie model map.
/// Output arguments: 'left', 'right' - the integer coordinates
/// of the lower left and of the upper right corner of the image cover.
template <class IntT, class Vect1, class Vect2, class IntvT>
inline int computeImage (const IntT *coord, int level,
	const Vect1 &theta, const Vect2 &p, const IntvT &lambda,
	IntT *left, IntT *right)
{
	// compute the actual box corresponding to the cube
	int intWidth = 1 << level;
	interval x [spaceDim];
	for (int i = 0; i < spaceDim; ++ i)
	{
		x [i] = interval (coord [i], coord [i] + 1);
		x [i] /= intWidth;
		x [i] *= width [i];
		x [i] += offset [i];
	}

	// compute the image of this box
	interval y [spaceDim];
	Leslie (x, theta, p, lambda, y, spaceDim);

	// rescale the image box to the integer coordinates and make sure
	// that the actual image is contained in the INTERIOR of the box
	for (int i = 0; i < spaceDim; ++ i)
	{
		y [i] -= offset [i];
		y [i] /= width [i];
		y [i] *= intWidth;
		double left_b = y [i]. leftBound ();
		left [i] = static_cast<IntT> (left_b);
		while (left [i] >= left_b) // use >= for the interior
			-- left [i];
		double right_b = y [i]. rightBound ();
		right [i] = static_cast<IntT> (right_b);
		while (right [i] <= right_b) // use <= for the interior
			++ right [i];
	}

	return 0;
} /* computeImage */

/// Computes the graph representation of the multivalued cubical enclosure
/// of the Leslie population model map on X.
template<class Vect1, class Vect2, class IntvT, class tSetOfCubes,
	class tDirectedGraph>
void computeMapGraph (const tSetOfCubes &X, int level,
	const Vect1 &theta, const Vect2 &p, const IntvT &lambda,
	tDirectedGraph &g)
{
	typedef typename tSetOfCubes::value_type tCube;
	typedef typename tCube::CoordType tCoord;
	using chomp::homology::tRectangle;
	using chomp::homology::sbug;

	// the sum of the numbers of cubes in the images of cubes
	int countimages = 0;

	// compute the images of all the cubes
	int countX = X. size ();
	for (int cur = 0; cur < countX; ++ cur)
	{
		// create the empty image
		g. addVertex ();

		// compute the image of the cube
		tCoord coord [spaceDim], left [spaceDim], right [spaceDim];
		X [cur]. coord (coord);
		computeImage (coord, level, theta, p, lambda, left, right);

		// add the corresponding edges to the graph
		tRectangle<tCoord> r (left, right, spaceDim);
		const tCoord *c;
		while ((c = r. get ()) != 0)
		{
			int number = X. getnumber (tCube (c, spaceDim));
			if (number < 0)
				continue;
			++ countimages;
			g. addEdge (number);
		}
	}

	// show the average size of the image of a cube
	if (sbug. show && !X. empty ())
		sbug << "[avg " << (countimages / X. size ()) << "] ";

	return;
} /* computeMapGraph */

template <class tSetOfCubes>
inline void subdivideCubes (const tSetOfCubes &X, tSetOfCubes &result)
{
	typedef typename tSetOfCubes::value_type tCube;
	typedef typename tCube::CoordType tCoord;
	using chomp::homology::tRectangle;

	if (X. empty ())
		return;
	int dim = X [0]. dim ();
	if (dim != spaceDim)
		return;

	tCoord cmin [spaceDim];
	tCoord cmax [spaceDim];
	int n = X. size ();
	for (int i = 0; i < n; ++ i)
	{
		X [i]. coord (cmin);
		for (int j = 0; j < dim; ++ j)
		{
			cmin [j] *= 2;
			cmax [j] = cmin [j] + 2;
		}
		tRectangle<tCoord> r (cmin, cmax, dim);
		const tCoord *c;
		while ((c = r. get ()) != 0)
		{
			tCube q (c, dim);
			result. add (q);
		}
	}
	return;
} /* subdivideCubes */

/// Computes a lower bound for the number of attractors for the Lesslie model
/// with the given parameters.
template<class Vect1, class Vect2, class IntvT>
int countAttractors (const Vect1 *theta, const Vect2 *p, const IntvT &lambda)
{
	typedef chomp::homology::tCubeFix<spaceDim,short int> tCube;
	typedef tCube::CoordType tCoord;
	typedef chomp::homology::hashedset<tCube> tSetOfCubes;
	typedef chomp::homology::diGraph<> tDirectedGraph;
	typedef chomp::homology::multitable<int> intTable;
	typedef chomp::homology::multitable<char> charTable;
	using chomp::homology::sbug;
	using chomp::homology::SCC;

	// prepare a single zero cube that represents the entire area R
	tCoord zeroArray [spaceDim];
	for (int i = 0; i < spaceDim; ++ i)
		zeroArray [i] = 0;
	tCube zeroCube (zeroArray, spaceDim);
	tSetOfCubes X;
	X. add (zeroCube);

	// subdivide the initial cube
	int level = 1;
	while (level < iniLevel)
	{
		tSetOfCubes Y;
		subdivideCubes (X, Y);
		X. swap (Y);
		++ level;
	}

	// subdivide the cubes and then compute the invariant part
	// or count the number of attractors
	while (1)
	{
		// subdivide the set of cubes into a higher subdivision level
		{
			tSetOfCubes Y;
			sbug << "Subdiv " << ((level < 10) ? "0" : "") <<
				level << ": ";
			subdivideCubes (X, Y);
			X. swap (Y);
			sbug << X. size () << ". ";
		}

		// compute the graph which represents the map on the cubes
		sbug << "F... ";
		tDirectedGraph g;
		computeMapGraph (X, level, theta, p, lambda, g);

		// prepare an array for marking vertices which remain
		// for the next run of the loop
		intTable selection;
		int nVertices = g. countVertices ();
		for (int i = 0; i < nVertices; ++ i)
			selection [i] = 0;

		// compute the chain recurrent components of the dynamics,
		// count the number of attractors, and determine the union
		// of these components and its image
		intTable compVertices;
		intTable compEnds;
		int nComponents = SCC (g, compVertices, compEnds);
		sbug << nComponents << " scc ";
		int nAttractors = 0;
		for (int n = 0; n < nComponents; ++ n)
		{
			// determine the vertex range of this component
			int beginVertex = n ? compEnds [n - 1] : 0;
			int endVertex = compEnds [n];

			// mark the entire component
			for (int i = beginVertex; i < endVertex; ++ i)
				selection [compVertices [i]] = n + 1;

			// mark the image of the component
			// and check if this component is an attractor
			bool attractor = true;
			for (int i = beginVertex; i < endVertex; ++ i)
			{
				int vertex = compVertices [i];
				int nEdges = g. countEdges (vertex);
				for (int j = 0; j < nEdges; ++ j)
				{
					int target = g. getEdge (vertex, j);
					if (!selection [target])
					{
						selection [target] = -1;
						attractor = false;
					}
					else if (selection [target] != n + 1)
						attractor = false;
				}
			}
			if (attractor)
				++ nAttractors;
		}

		// indicate the computed number of attractors
		sbug << "(" << nAttractors << " attr) ";

		// if there is more than one attractor then return the answer
		if (nAttractors > 1)
		{
			sbug << "Good!\n";
			return nAttractors;
		}

		// if this is the last subdivision level but there is at most
		// one attractor then return the answer, although it is bad
		if (level == subLevel)
		{
			sbug << "Bad.\n";
			return nAttractors;
		}

		// restrict the set X to the computed chain recurrent set
		// and its cubical image (in order to detect non-attractors)
		tSetOfCubes Y;
		for (int i = 0; i < nVertices; ++ i)
		{
			if (selection [i])
				Y. add (X [i]);
		}
		X. swap (Y);
		sbug << X. size () << " left.\n";
		++ level;
	}
} /* computeAttractors */

/// This is the procedure which performs one computation for the given box.
/// Returns: >0 if the property holds true, 0 if this verification could not
/// prove that the property holds true, <0 if an error occurred.
int computeFunction (const double *left, const double *right,
	int dim, int level)
{
	using chomp::homology::sbug;

	// exclude the obvious cases
	if (dim != paramDim)
		return -1;
	if (!level)
		return 0;

	// prepare the parameters of the dynamical system
	interval theta [spaceDim];
	for (int i = 0; i < spaceDim; ++ i)
		theta [i] = interval (left [i], right [i]);
	interval p [spaceDim - 1];
	for (int i = 0; i < spaceDim - 1; ++ i)
		p [i] = interval (7, 7) / 10;
	interval lambda = interval (1, 1) / 10;

	// show debugging information what is being computed if necessary
	if (sbug. show)
	{
		sbug << "Counting attractors at level " << level << ":\n";
		sbug << "theta = ";
		for (int i = 0; i < spaceDim; ++ i)
		{
			sbug << (i ? " x [" : "[") << left [i];
			if (left != right)
				sbug << "," << right [i];
			sbug << "]";
		}
		sbug << ", p = ";
		for (int i = 0; i < spaceDim - 1; ++ i)
			sbug << (i ? " x " : "") << p [i]. leftBound ();
		sbug << ", l = " << lambda. leftBound () << ".\n";
	}

	// compute the number of attractors for the given dynamical system
	int n = countAttractors (theta, p, lambda);
	if (sbug. show)
	{
		sbug << n << ((n == 1) ? " attractor.\n" : " attractors.\n");
	}

	// return the number of attractors if greater than 1 or 0 otherwise
	return (n > 1) ? n : 0;

} /* computeFunction */


// --------------------------------------------------
// ---------------------- MAIN ----------------------
// --------------------------------------------------

int main (int argc, char *argv [])
{
	// call the actual "main" function
	return chomp::multiwork::mwSubdivMain<paramDim,int> (argc, argv,
		title, helpinfo, defaultPortNumber, controlNumber,
		leftBounds, rightBounds, minSubdivLevel, computeFunction);
} /* main */

